"""
Base interface for decoders
"""

import ast
from collections import namedtuple
import re
import logging
import os

import kordesii

Source = namedtuple("Source", ("name", "path"))


logger = logging.getLogger(__name__)


class Decoder(object):
    def __init__(self, script_path, name=None, source=None):
        """
        Initializes Decoder object.
        :param script_path: Path to script to run within dissasembler.
        :param name: Name of decoder, file basename is used it not provided.
        :param source: Optional Source namedtuple containing (source_name, source_path)
        """
        self._doc_string = None
        self._author = None
        self._description = None
        self.script_path = script_path
        # Just use directory name of script if source not provided.
        directory = os.path.dirname(script_path)
        self.source = source or Source(directory, directory)
        if name:
            self.name = name
        else:
            self.name, _ = os.path.splitext(os.path.basename(script_path))
            # Remove legacy postfix if it exists.
            if self.name.endswith("_StringDecode"):
                self.name = self.name[: -len("_StringDecode")]
        self.full_name = "{}:{}".format(self.source.name, self.name)

    def __repr__(self):
        return "<{} Decoder at {}>".format(self.name, self.script_path)

    @property
    def docstring(self):
        """Retrieves the docstring within the decoder source or '' if no docstring exists."""
        if self._doc_string is not None:
            return self._doc_string
        # Open up script file and parse for docstring.
        # (We can't import because these decoders are designed to work within a disassembler and
        #  will explode if we try to import them on the outside.)
        with open(self.script_path, "r") as fo:
            mod = ast.parse(fo.read(), self.script_path)
            try:
                self._doc_string = ast.get_docstring(mod)
            except TypeError:
                self._doc_string = ""
        return self._doc_string

    # TODO: Use the same "@" syntax used by ghidra.
    @property
    def description(self):
        """Retrieves description from docstring or '' if not found."""
        if self._description is not None:
            return self._description
        docstring = self.docstring
        if not docstring:
            self._description = ""
            return self._description
        match = re.search("Description:(.*?)$", docstring, re.MULTILINE)
        if match:
            self._description = match.group(1).strip()
        else:
            self._description = ""
        return self._description

    @property
    def author(self):
        """Retrieves author from docstring or '' if not found."""
        if self._author is not None:
            return self._author
        docstring = self.docstring
        if not docstring:
            self._author = ""
            return self._author
        match = re.search("Author:(.*?)$", docstring, re.MULTILINE)
        if match:
            self._author = match.group(1).strip()
        else:
            self._author = ""
        return self._author

    def run(self, input_file, reporter, **run_config):
        """
        Runs decoder on given input_file.

        :param input_file: File path to input file to run decoder on.
        :param reporter: Reporter object to use for reporting metadata.
        :param run_config: Run configuration to pass along to kordesii.run_ida()
        """
        # IDA doesn't like backslashes in it's argv.
        input_file = input_file.replace("\\", "/").strip()
        script_path = self.script_path.replace("\\", "/").strip()

        kordesii.run_ida(reporter, script_path, input_file, **run_config)



"""
Contains various helper functions for our tracing, which is static pseudo-execution of assembly.
"""

from builtins import bytes
from builtins import range
from builtins import object
import logging

import idc
import idaapi

logger = logging.getLogger(__name__)

# FE
POS_FIRST = 0
POS_SECOND = 1
POS_THIRD = 2

ESI_REG_FAM = ["rsi", "esi", "si", "sil"]
EDI_REG_FAM = ["rdi", "edi", "di", "dil"]
EBP_REG_FAM = ["rbp", "ebp", "bp", "bpl"]
ESP_REG_FAM = ["rsp", "esp", "sp", "spl"]
EIP_REG_FAM = ["rip", "eip", "ip"]
EAX_REG_FAM = ["rax", "eax", "ax", "ah", "al"]
EBX_REG_FAM = ["rbx", "ebx", "bx", "bh", "bl"]
ECX_REG_FAM = ["rcx", "ecx", "cx", "ch", "cl"]
EDX_REG_FAM = ["rdx", "edx", "dx", "dh", "dl"]
R8_REG_FAM = ["r8", "r8d", "r8w", "r8b"]
R9_REG_FAM = ["r9", "r9d", "r9w", "r9b"]
R10_REG_FAM = ["r10", "r10d", "r10w", "r10b"]
R11_REG_FAM = ["r11", "r11d", "r11w", "r11b"]
R12_REG_FAM = ["r12", "r12d", "r12w", "r12b"]
R13_REG_FAM = ["r13", "r13d", "r13w", "r13b"]
R14_REG_FAM = ["r14", "r14d", "r14w", "r14b"]
R15_REG_FAM = ["r15", "r15d", "r15w", "r15b"]
XMM0_REG_FAM = ["xmm0"]
XMM1_REG_FAM = ["xmm1"]

""" All x86 registers, broken up into families (such as rax, eax, ax, ah, al)"""
REG_FAM = [
    ESI_REG_FAM,
    EDI_REG_FAM,
    EBP_REG_FAM,
    ESP_REG_FAM,
    EIP_REG_FAM,
    EAX_REG_FAM,
    EBX_REG_FAM,
    ECX_REG_FAM,
    EDX_REG_FAM,
    R8_REG_FAM,
    R9_REG_FAM,
    R10_REG_FAM,
    R11_REG_FAM,
    R12_REG_FAM,
    R13_REG_FAM,
    R14_REG_FAM,
    R15_REG_FAM,
    XMM0_REG_FAM,
    XMM1_REG_FAM,
]

""" Non-general registers """
RESERVED_REG_FAM = [ESI_REG_FAM, EDI_REG_FAM, EBP_REG_FAM, ESP_REG_FAM, EIP_REG_FAM]

""" Move mnemonics """
MOVS = ("rep mov", "movsb", "movsw", "movsd")


class TraceState(object):
    """
    An object to hold all tracing's stateful info.
    """

    def __init__(self):
        self.regs = {}
        self.stack = {}
        self.visited_eas = []
        self.pp_track = []  # push/pop tracker

    def get_reg_value(self, reg):
        """Get the value of the register. The register or the register family can be specified.

        :param reg: The register to retrieve the value of. May be the register family.

        :return: The value of the register if known, or None
        """
        if isinstance(reg, (str, bytes)):
            reg = unsafe_get_reg_fam(reg)
        if reg:
            return self.regs.get(reg[0], (None, None))[0]
        else:
            return None

    def set_reg_value(self, reg, value, ea):
        """Set the value of the register. The register or the register family can be specified.

        :param reg: The register to set the value of. May be the register family.
        :param value: The value to set, or None if the value is not known/cannot be determined.
        :param ea: The address where the register is being set.

        :return:
        """
        if isinstance(reg, (str, bytes)):
            reg = unsafe_get_reg_fam(reg)
        if reg:
            if value is not None:
                self.regs[reg[0]] = (value, ea)
            elif reg[0] in self.regs:
                del self.regs[reg[0]]


class BranchingTraceState(TraceState):
    """
    An object to hold all tracing's stateful info including ea.
    """

    def __init__(self, ea=None, state=None):
        super(BranchingTraceState, self).__init__()
        self.ea = idc.BADADDR if ea is None else ea
        if state is not None:
            if ea == idc.BADADDR:
                self.ea = state.ea
            # Note: this needs to be "global" amongst all derived states, so it cannot be a new list.
            self.visited_eas = state.visited_eas
            self.pp_track = list(state.pp_track)
            self.regs = dict(state.regs)
            self.stack = dict(state.stack)


def unsafe_get_reg_fam(reg):
    """Gets the register family for any register

        :param reg: register string representation such as 'eax'

        :return: list of all associated register names, such as ['rax', 'eax',  'ax', 'ah', 'al']
    """
    for family in REG_FAM:
        if reg in family:
            return family


def get_reg_fam(reg):
    """ Gets the register family, returns None if it is a reserved register

        :param reg: register string representation such as 'eax'

        :return: list of all associated register names, such as ['rax', 'eax',  'ax', 'ah', 'al']
    """
    if get_reserved_reg(reg):
        return
    return unsafe_get_reg_fam(reg)


def get_reserved_reg(reg):
    """Gets the family of a reserved register, returns None otherwise

        :param reg: register string representation such as 'eax'

        :return: list of all associated register names, such as ['rax', 'eax',  'ax', 'ah', 'al']
    """
    for family in RESERVED_REG_FAM:
        if reg in family:
            return family


def get_reg_value(reg_values, reg):
    """
        Helper function to retrieve the current value of a register given the register dictionary returned from
        create_stack and the string representation of the register name

        :param reg_values: register dictionary returned by create_stack function
        :param reg: string representation of the register name

        :return: value of register, None if register not in dictionary
    """
    regfam = unsafe_get_reg_fam(reg)
    if regfam:
        try:
            if reg_values[regfam[0]]:
                return reg_values[regfam[0]][0]
        except KeyError:
            return None
    else:
        return None


def get_reg_value_offset(reg_values, reg):
    """
    Obtain the tuple of value, offset for a specified register when provided the TracerState().regs dictionary.

    :param reg_values: TracerState().regs dictionary
    :param reg: Target register

    :return: Tuple of value, offset if register in dictionary, otherwise None, None
    """
    regfam = unsafe_get_reg_fam(reg)
    if regfam:
        try:
            return reg_values[regfam[0]]
        except KeyError:
            return None, None
    else:
        return None, None


def is_64_bit():
    """Returns True if the code is 64-bit code, false otherwise"""
    return idaapi.get_inf_structure().is_64bit()


def get_int(val):
    """Attempts to convert the string given into the correct integer"""
    if val.isdigit():
        return int(val)
    elif val.endswith("h"):
        return int(val[:-1], 16)
    elif val.endswith("o"):
        return int(val[:-1], 8)
    elif val.endswith("b"):
        return int(val[:-1], 2)


OPERAND_BYTE_SIZES = {
    0: 1,
    1: 2,
    2: 4,
    3: 4,
    4: 8,
    5: 4,  # Actually variable, so may not be 4
    6: 4,  # Actually packed real format for mc68040, so may not be 4
    7: 8,
    8: 16,
    9: 4,  # Actually a pointer, so may not be 4
    10: 4,  # Actually a void, so may not be 4
    11: 5,
    12: 1,  # Actually a bit field, so less than one.
    13: 4,  # Actually a pointer, so may not be 4
    14: 4,  # Actually a pointer, so may not be 4
    15: 3,
    16: 16,  # Actually a long double, so may not be 16
    17: 32,
    18: 64,
}


def get_byte_size_of_operand(ea, pos):
    """Gets the byte size of the operand at the given ea and position"""
    cmd = idaapi.insn_t()
    idaapi.decode_insn(cmd, ea)
    return OPERAND_BYTE_SIZES.get(cmd.ops[pos].dtype, 4)  # 4 is Unknown


def get_opnd_replacement(ea, pos):
    """ A replacement for IDA's idc.print_operand that can de-alias register names"""
    # TODO: Support renames in other operand types
    if idc.get_operand_type(ea, pos) == idc.o_reg:
        return idaapi.get_reg_name(idc.get_operand_value(ea, pos), get_byte_size_of_operand(ea, pos))
    else:
        return idc.print_operand(ea, pos)


def obtain_phrase_register(ea, pos):
    """
    Obtain the register referenced in an idc.o_phrase

    :param ea: Memory location
    :param pos: Argument location

    :return: Register or None
    """
    if idc.get_operand_type(ea, pos) == idc.o_phrase:
        opnd = get_opnd_replacement(ea, pos)
        for family in REG_FAM:
            for member in family:
                if member in opnd:
                    return member


def is_displ(ea, pos):
    """Determines if the ea/position is a displacement, such as [esp + 8]

        :param ea: memory location
        :param pos: argument location
                    example:
                        add eax, ebx
                        eax is position 0, ebx is position 1

        :return: True or False
    """
    return idc.get_operand_type(ea, pos) in (idc.o_phrase, idc.o_displ)


def get_operand_value_replacement(ea, pos, state):
    """ A replacement for Ida's idc.get_operand_value that handles displacements more reasonably

        :param ea: memory location
        :param pos: argument location
                    example:
                        add eax, ebx
                        eax is position 0, ebx is position 1
        :param state: the current stack pointer register family (usually sp)

        :return: computes a numerical replacement for an operand
    """
    if is_displ(ea, pos):
        bit_size = 64 if is_64_bit() else 32
        stack_reg = "rsp" if bit_size == 64 else "esp"
        cmd = idaapi.insn_t()
        idaapi.decode_insn(cmd, ea)
        offset = cmd.ops[pos].addr
        # Convert the offset to a signed value
        if offset & (1 << (bit_size - 1)):
            offset -= 1 << bit_size
        if stack_reg in get_opnd_replacement(ea, pos):
            offset += idc.get_spd(ea) or 0
        return offset
    else:
        return idc.get_operand_value(ea, pos)


def get_encoded_stack_string(stack, startoffset, size=None):
    """
        given a stack and the starting offset, pull out the string from the stack\
        if size is given, pull that many bytes off stack

        :param stack: stack dictionary of your current working stack as created by the create_stack function
        :param startoffset: offset to start from on stack
        :param size: optional parameter for size to pull off stack, if not set will read until a null byte

        :return: string of characters from the stack dictionary starting at offset and reading until size or a null byte
    """
    encrypted = []
    offset = startoffset
    while stack.get(offset):
        if not size and stack[offset][0] == 0:
            break
        if size and size + startoffset == offset:
            break
        encrypted.append(stack[offset][0])
        offset += 1
    return bytes(encrypted)


def get_encoded_stack_string_wide(stack, startoffset, size=None):
    """
        given a stack and the starting offset, pull out the string from the stack\
        if size is given, pull that many bytes off stack

        :param stack: stack dictionary of your current working stack as created by the create_stack function
        :param startoffset: offset to start from on stack
        :param size: optional parameter for size to pull off stack, if not set will read until two null bytes
                    (wchar null)

        :return: string of characters from the stack dictionary starting at offset and reading until size or a null byte
                 this will not be wide characters
    """
    encrypted = []
    offset = startoffset
    while stack.get(offset) and stack.get(offset + 1):
        if not size and stack[offset][0] == 0 and stack[offset + 1][0] == 0:
            break
        if size and size + startoffset >= offset:
            break
        encrypted.append(stack[offset][0])
        encrypted.append(stack[offset + 1][0])
        offset += 2
    return bytes(encrypted)


def set_stack(offset, ea, pos, state):
    """Sets the stack dictionary, at the given offset, to contain the value at the given position at the given ea,
        performing a lookup in the register dictionary if needed. Used by create_stack

        :param offset: offset to set on stack
        :param ea: instruction location
        :param pos: argument position
        :param state: the current TraceState

        :return: None - updates state
    """
    if idc.get_operand_type(ea, pos) == idc.o_imm:
        val = idc.get_operand_value(ea, pos)
    else:
        val = state.get_reg_value(get_opnd_replacement(ea, pos))
    if val is not None:
        for i in range(get_byte_size_of_operand(ea, pos)):
            state.stack[offset + i] = (val & 0xFF, ea)
            val >>= 8


def handle_string_mov(ea, state):
    """Updates the stack based on a movs instruction.  Used by create_stack
        If a rep/repne prefix is used, takes the count from ecx.  If the count cannot be determined, will ignore
        the instruction.  Also assumes that esi points to memory within the executable, and edi points to the
        stack. On any errors, this will ignore the instruction.

        :param ea: instruction location
        :param state: the current TraceState

        :return: None - updates stack or regs
    """
    opcode = idaapi.get_bytes(ea, 1)
    rep_inst = opcode in [b"\xf2", b"\xf3"]
    count = state.get_reg_value("ecx") if rep_inst else 1
    if not count or count < 0:
        return

    cmd = idaapi.insn_t()
    inslen = idaapi.decode_insn(cmd, ea)
    dtype = cmd.ops[0].dtype
    word_size = [1, 2, 4][dtype] if dtype < 3 else 4
    count *= word_size

    src = state.get_reg_value("esi")
    dst = state.get_reg_value("edi")
    if src is None or dst is None:
        return
    # In IDA 7, get_bytes doesn't return None on failure, instead it will return
    # a string of \xff the size of count. My theory is that the function changed
    # to return -1 for each byte within the c code and something is casting it to a string before returning.
    # Since, all \xff's could be valid we need to check if src is valid instead.
    if not idc.is_loaded(src):
        return
    bytes_ = idaapi.get_bytes(src, count)
    if bytes_ in (None, -1):  # Keep this around in-case they fix it in a future version.
        return
    for i in range(count):
        state.stack[dst + i] = ((bytes_[i]), ea)

    if rep_inst:
        state.set_reg_value("ecx", 0, ea)
    state.set_reg_value("esi", src + count, ea)
    state.set_reg_value("edi", dst + count, ea)


def handle_mov(ea, state):
    """Updates the stack based on a mov instruction. Used by create_stack

        :param ea: instruction location
        :param state: the current TraceState

        :return: None - updates stack or regs
    """
    op1 = get_opnd_replacement(ea, POS_FIRST)
    if idc.get_operand_type(ea, POS_FIRST) != idc.o_reg:
        offset = get_operand_value_replacement(ea, POS_FIRST, state)
        set_stack(offset, ea, POS_SECOND, state)
    else:
        type_ = idc.get_operand_type(ea, POS_SECOND)
        val = None
        if type_ == idc.o_reg:
            val = state.get_reg_value(get_opnd_replacement(ea, POS_SECOND))
        elif type_ == idc.o_mem:
            bytes = idc.get_bytes(idc.get_operand_value(ea, POS_SECOND), get_byte_size_of_operand(ea, POS_SECOND))
            if bytes:
                val = 0
                for x in range(len(bytes)):
                    val += (bytes[x]) << (8 * x)
        elif type_ == idc.o_imm:
            val = idc.get_operand_value(ea, POS_SECOND)
        else:
            offset = get_operand_value_replacement(ea, POS_SECOND, state)
            val, ea = state.stack.get(offset, (None, ea))
        state.set_reg_value(op1, val, ea)


def handle_lea(ea, state):
    """Updates the stack based on an lea instruction. Used by create_stack

        :param ea: instruction location
        :param state: the current TraceState

        :return: None - updates stack or regs
    """
    value = get_operand_value_replacement(ea, POS_SECOND, state)
    if not value and idc.get_operand_type(ea, POS_SECOND) != idc.o_mem:
        return

    state.set_reg_value(get_opnd_replacement(ea, POS_FIRST), value, ea)


def handle_push(ea, state):
    """
        Loosely tracks the values that get pushed onto the stack, if the value is unknown when pushed.
        Should be called any time you see a push instruction if tracking

        :param ea: instruction location
        :param state: the current TraceState
    """
    op_type = idc.get_operand_type(ea, POS_FIRST)
    if op_type == idc.o_reg:
        value = state.get_reg_value(get_opnd_replacement(ea, POS_FIRST))
    elif op_type == idc.o_imm:
        value = get_operand_value_replacement(ea, POS_FIRST, state)
    else:
        value = None
    state.pp_track.append(value)


def handle_pop(ea, state):
    """
        if op type is a reg, pops the tracked value if not none into the appropriate reg
        should be called anytime a pop is seen if tracking

        :param ea: instruction location
        :param state: the current TraceState
    """
    value = state.pp_track.pop() if state.pp_track else None
    if idc.get_operand_type(ea, POS_FIRST) == idc.o_reg:
        state.set_reg_value(get_opnd_replacement(ea, POS_FIRST), value, ea)


def handle_test(ea, state):
    """
        If a test of a register against itself occurs and the next instruction is a jnz,
        then the register can be set to zero (code is followed linearly, jumps are ignored),
        unless the next instruction is a jmp.

        :param ea: instruction location
        :param state: the current TraceState
    """
    if idc.get_operand_type(ea, POS_FIRST) == idc.o_reg and idc.get_operand_type(ea, POS_SECOND) == idc.o_reg:
        op1 = get_opnd_replacement(ea, POS_FIRST)
        op2 = get_opnd_replacement(ea, POS_SECOND)
        next_ea = ea + idc.get_item_size(ea)
        if op1 == op2 and idc.print_insn_mnem(next_ea) == "jnz":
            next_ea += idc.get_item_size(next_ea)
            if not idc.print_insn_mnem(next_ea).startswith("j"):
                state.set_reg_value(op1, 0, ea)


def create_state(end_ea, start_ea=None):
    """
        Quick and dirty representation of stack and regs from start of function to this ea.

        :param end_ea: The EA of which you want to compute the stack up until
        :param start_ea: Optional param of the beginning of the function - sometimes necessary if ida can't
                        compute and you can

        :return A newly created TraceState
    """
    if not start_ea:
        start_ea = idaapi.get_func(end_ea).start_ea
    state = TraceState()
    ea = start_ea
    while ea < end_ea:
        mnemonic = idc.print_insn_mnem(ea)
        if mnemonic.startswith("movs"):
            handle_string_mov(ea, state)
        elif mnemonic.startswith("mov"):
            handle_mov(ea, state)
        elif mnemonic in ["xor", "sub"] and get_opnd_replacement(ea, POS_FIRST) == get_opnd_replacement(ea, POS_SECOND):
            state.set_reg_value(get_opnd_replacement(ea, POS_FIRST), 0, ea)
        elif mnemonic == "lea":
            handle_lea(ea, state)
        elif mnemonic == "push":
            handle_push(ea, state)
        elif mnemonic == "pop":
            handle_pop(ea, state)
        elif mnemonic in ["test", "cmp"]:
            handle_test(ea, state)
        elif mnemonic:
            if idc.get_operand_type(ea, POS_FIRST) == idc.o_reg:
                state.set_reg_value(get_opnd_replacement(ea, POS_FIRST), None, ea)

        ea += idc.get_item_size(ea)
    return state


def trace_rep_mov(stack_var, loc, func_ea, state):
    """
    Helper function to trace back a rep mov
    """
    loc = idc.prev_head(loc)
    while loc != func_ea:
        mnem = idc.print_insn_mnem(loc)
        if mnem == "lea" and idc.print_operand(loc, 0) == "edi":
            if stack_var == get_operand_value_replacement(loc, 1, state):
                return trace_register("esi", loc, func_ea, state)
            return None
        loc = idc.prev_head(loc)
    return None


def trace_stack_var(stack_var, loc, func_ea, state=None):
    """
    Trace a provided stack variable to the location in which it is set. If it is set using a mov operation, the
    value may be set from a register, which we then need to call back to trace_register in order to acquire the
    value. Otherwise it is either an o_imm type (return the immediate value) or an o_mem type, that means the
    referenced location MAY a pointer to the actual data, which we need to acquire. We validate this by ensuring
    the acquired value is within the loaded memory range of the application. Otherwise we return idc.BADADDR.

    :param stack_var: The stack variable which a location is loaded into
    :param loc: The starting offset for tracing back from
    :param func_ea: Starting function offset
    :param state: the current TraceState

    :return: The acquired location, or idc.BADADDR
    """
    if state is None:
        state = TraceState()
    loc = idc.prev_head(loc)
    while loc != func_ea:
        op_type_1 = idc.get_operand_type(loc, 1)
        opval_1 = idc.get_operand_value(loc, 1)
        mnem = idc.print_insn_mnem(loc)
        dis = idc.GetDisasm(loc)
        if mnem == "mov" and stack_var == get_operand_value_replacement(loc, 0, state):
            if op_type_1 == idc.o_reg:
                return trace_register(idc.print_operand(loc, 1), loc, func_ea, state)
            elif op_type_1 == idc.o_imm:
                return opval_1
            elif op_type_1 == idc.o_mem:
                poss_loc = idc.get_wide_dword(opval_1)
                if idaapi.cvar.inf.minEA <= poss_loc < idaapi.cvar.inf.maxEA:
                    return poss_loc
                else:
                    return opval_1
            else:
                return idc.BADADDR
        elif any(x in dis for x in MOVS):
            result = trace_rep_mov(stack_var, loc, func_ea, state)
            if result:
                return result
        loc = idc.prev_head(loc)
    return idc.BADADDR


def trace_register(reg, loc, func_ea, state=None):
    """
    Trace a provided register to the location in which it is set. If it is set using a mov operation, the
    value is either an o_imm type (return the immediate value) or an o_mem type, that means the referenced location
    MAY a pointer to the actual data, which we need to acquire. We validate this by ensuring the acquired value is
    within the loaded memory range of the application. Otherwise we return idc.BADADDR.

    If the register is set using an lea operation, it is likely done so using a stack variable (which we validate)
    and then trace back to determine how the stack variable is set.

    :param reg: The referenced register which a location is loaded into
    :param loc: The starting offset for tracing back from
    :param func_ea: Starting function offset
    :param state: the current TraceState

    :return: The acquired location, or idc.BADADDR
    """
    if state is None:
        state = TraceState()
    loc = idc.prev_head(loc)
    while loc != func_ea:
        opnd = idc.print_operand(loc, 0)
        op_type_1 = idc.get_operand_type(loc, 1)
        opval_1 = idc.get_operand_value(loc, 1)
        mnem = idc.print_insn_mnem(loc)
        if "mov" in mnem and opnd == reg:
            if op_type_1 == idc.o_imm:
                return opval_1
            elif op_type_1 == idc.o_mem:
                poss_loc = idc.get_wide_dword(opval_1)
                if idaapi.cvar.inf.minEA <= poss_loc < idaapi.cvar.inf.maxEA:
                    return poss_loc
                else:
                    return opval_1
            else:
                return idc.BADADDR
        elif mnem == "lea" and op_type_1 == idc.o_displ and opnd == reg:
            stack_var = get_operand_value_replacement(loc, 1, state)
            return trace_stack_var(stack_var, loc, func_ea, state)
        elif is_64_bit() and mnem == "lea" and opnd == reg:
            if op_type_1 == idc.o_mem:
                return opval_1
        loc = idc.prev_head(loc)
    return idc.BADADDR


def trace_register_family(reg, loc, func_ea, state=None):
    """
    Trace a provided register to the location in which any member of its register family is set. If it is set using a
    mov operation, the value is either an o_imm type (return the immediate value) or an o_mem type, that means the
    referenced location MAY a pointer to the actual data, which we need to acquire. We validate this by ensuring the
    acquired value is within the loaded memory range of the application. Otherwise we return idc.BADADDR.

    If the register is set using an lea operation, it is likely done so using a stack variable (which we validate)
    and then trace back to determine how the stack variable is set.

    :param reg: The referenced register which a location is loaded into
    :param loc: The starting offset for tracing back from
    :param func_ea: Starting function offset
    :param state: the current TraceState (a new state will be created if one is not provided)

    :return: The acquired location, or idc.BADADDR
    """
    if state is None:
        state = TraceState()
    reg_fam = unsafe_get_reg_fam(reg)
    if reg_fam:
        loc = idc.prev_head(loc)
        while loc != func_ea:
            opnd = idc.print_operand(loc, 0)
            op_type_1 = idc.get_operand_type(loc, 1)
            opval_1 = idc.get_operand_value(loc, 1)
            mnem = idc.print_insn_mnem(loc)
            if "mov" in mnem and opnd in reg_fam:
                if op_type_1 == idc.o_imm:
                    return opval_1
                elif op_type_1 == idc.o_mem:
                    poss_loc = idc.get_wide_dword(opval_1)
                    if idaapi.cvar.inf.minEA <= poss_loc < idaapi.cvar.inf.maxEA:
                        return poss_loc
                    else:
                        return opval_1
                else:
                    return idc.BADADDR
            elif mnem == "lea" and op_type_1 == idc.o_displ and opnd in reg_fam:
                stack_var = get_operand_value_replacement(loc, 1, state)
                return trace_stack_var(stack_var, loc, func_ea, state)
            elif is_64_bit() and mnem == "lea" and opnd in reg_fam:
                if op_type_1 == idc.o_mem:
                    return opval_1
            loc = idc.prev_head(loc)
    return idc.BADADDR


def trace_register_family_x64(reg, loc, func_ea, state=None):
    """
    Shouldn't be any different than trace_register_family except for continuing a trace if a register is loaded into
    the target register. However, using it in trace_register_family was not behaving properly, so keeping it separate
    for now. Also unsure of what the repercussions would be of adding that into the existing trace_register_family.

    :param reg: The referenced register which a location is loaded into
    :param loc: The starting offset for tracing back from
    :param func_ea: Starting function offset
    :param state: the current TraceState

    :return: The acquired location, or idc.BADADDR
    """
    if state is None:
        state = TraceState()
    reg_fam = unsafe_get_reg_fam(reg)
    if reg_fam:
        loc = idc.prev_head(loc)
        while loc != func_ea:
            opnd = idc.print_operand(loc, 0)
            op_type_1 = idc.get_operand_type(loc, 1)
            opval_1 = idc.get_operand_value(loc, 1)
            mnem = idc.print_insn_mnem(loc)
            if "mov" in mnem and opnd in reg_fam:
                if op_type_1 == idc.o_imm:
                    return opval_1
                elif op_type_1 == idc.o_mem:
                    poss_loc = idc.get_wide_dword(opval_1)
                    if idaapi.cvar.inf.minEA <= poss_loc < idaapi.cvar.inf.maxEA:
                        return poss_loc
                    else:
                        return opval_1
                elif op_type_1 == idc.o_reg:
                    return trace_register_family_x64(idc.print_operand(loc, 1), loc, func_ea, state)
                else:
                    return idc.BADADDR
            elif mnem == "lea" and op_type_1 == idc.o_displ and opnd in reg_fam:
                stack_var = get_operand_value_replacement(loc, 1, state)
                return trace_stack_var(stack_var, loc, func_ea, state)
            elif is_64_bit() and mnem == "lea" and opnd in reg_fam:
                if op_type_1 == idc.o_mem:
                    return opval_1
            loc = idc.prev_head(loc)
    return idc.BADADDR



import functools
import inspect
import logging
import os
from io import open

from kordesii.core import DECODER_OUTPUT_DIR
from ruamel.yaml import YAML

yaml = YAML(typ="safe")
yaml.default_flow_style = False
log = logging.getLogger(__name__)

_serializers = {}


class ROObject(object):
    """Basic class represent deserialized objects that are read-only."""

    def __init__(self, adict):
        self.__dict__.update(adict)

    def __repr__(self):
        if "repr" in self.__dict__:
            return self.__dict__["repr"]

        repr_str = "<{}.{}<{}> at {:#8x}>".format(
            self.__module__, self.__class__.__name__, self.__dict__.get("__yaml_name__"), id(self)
        )

        return repr_str

    def __setattr__(self, key, value):
        raise RuntimeError("Cannot modify attributes of Read Only Objects.")

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(dict(constructor.construct_pairs(node)))


yaml.constructor.add_constructor(u"!ReadOnlyObject", ROObject.from_yaml)


def obj_to_yaml(representer, node, yaml_name=None, skip_attrs=None):
    """
    Generic YAML representer function. Converts the arbitrary class instance to a
    ReadOnlyObject formatted YAML representation.

    By default this captures every non-private, non-function attribute of the class instance.
    This includes any properties set with ``@property`` or ``property()``.

    :param ruamel.yaml.representer.BaseRepresenter representer: YAML representer
    :param node: Class instance to represent in YAML
    :param yaml_name: Optional tag for YAML serialization
    :param skip_attrs: Attributes to skip serializing
    :return: YAML representation object for node
    """
    skip_attrs = skip_attrs or []
    members = {}
    # Adapted from inspect.getmembers()
    for key in dir(node):
        if key.startswith("_") or key in skip_attrs:
            continue
        try:
            value = getattr(node, key)
        except AttributeError:
            continue
        if not inspect.isroutine(value):
            members[key] = value

    if "__yaml_name__" not in members:
        yaml_name = yaml_name or node.__class__.__name__
        members["__yaml_name__"] = yaml_name

    return representer.represent_mapping(u"!ReadOnlyObject", members)


def serializable_class(cls=None, yaml_name=None, skip_attrs=None, func=None):
    """
    Class decorator for arbitrary YAML-serializable classes.

    Optionally takes a function as an argument to overwrite the generic
    `obj_to_yaml` or the class's `to_yaml` function.
    This function must also be a valid YAML representer function.

    If a function is not specified, the class function `to_yaml` will be used
    if it exists. Otherwise, the generic `obj_to_yaml` will be used.
    """

    # A wrapper is needed for the case of using arguments in the decorator
    def _wrapper(klass):
        yaml_func = func  # separate variable to avoid strange wrapper reference issues
        if yaml_func is None:
            if hasattr(klass, "to_yaml"):
                yaml_func = klass.to_yaml
            elif yaml_name is None and skip_attrs is None:
                yaml_func = obj_to_yaml
            else:
                yaml_func = functools.partial(obj_to_yaml, yaml_name=yaml_name, skip_attrs=skip_attrs)
        yaml.representer.add_representer(klass, yaml_func)
        return klass

    if cls is None:
        return _wrapper

    return _wrapper(cls)


class Serializer(object):
    """
    Do not access directly, instead, access via ``get_serializer()``.

    Serialization object. To be used within decoders to save arbitrary
    data that can later be used in an MWCP parser or elsewhere.

    Serialized data is saved to the provided file path in yaml.

    A Serializer instance acts somewhat similarly to a dictionary, with the caveat that
    all keys are **write once**.

    Note: data is serialized and saved **when set**, so if a class is stateful,
    the version at the time of the key referring to that instance
    being set is the version saved, even if it is
    modified later. This also means that if that instance is retrieved from the Serializer
    later, it may not be the same version as what was saved. The best practice is
    to only serialize an instance when it will no longer be modified (e.g. right before
    the decoder is finished.) This should only be a real concern for EncodedString
    and EncodedStackString.
    """

    def __init__(self, file_path):
        self._data = {}
        self._file = None

        # Normally the directory hasn't been created yet when the Serializer is
        # first initialized.
        directory = os.path.dirname(file_path)
        if not os.path.isdir(directory):
            os.makedirs(directory)

        self._filepath = file_path
        self._name, _ = os.path.splitext(os.path.basename(file_path))

    def __del__(self):
        """
        Close the file when the Serializer object is deleted or garbage collected
        (including termination of the decoder).
        """
        self.close()

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __contains__(self, item):
        return item in self._data

    def __repr__(self):
        return "<Serializer with keys: {!r}>".format(self._data.keys())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # NOTE: We can't close because this class gets used as a global variable.
        return False

    def close(self):
        """Closes the Serializer object."""
        if self._file is not None and not self._file.closed:
            self._file.close()

    def keys(self):
        """List of all keys serialized."""
        return self._data.keys()

    def values(self):
        """List of all values serialized."""
        return self._data.values()

    def items(self):
        """Tuple of (key, value) all pairs serialized."""
        return self._data.items()

    def as_dict(self):
        """Return current serialized data as a dictionary."""
        return self._data.copy()

    def get(self, key, default=None):
        """Retrieve a previously saved key. Accepts a default."""
        return self._data.get(key, default)

    def set(self, key, value):
        """
        Set a key-value pair and serialize them.

        Keys may only be set once, a `ValueError` will be raised
        if attempting to set a previously set key.

        The pair is saved to the serialization file immediately.

        :param key: Any YAML-valid key, like a string or number
        :param value: Arbitrary value to serialize, must be YAML compatible
        :raises ValueError: A used key is attempted to be used again
        """
        if key in self._data:
            log.error("Key '{}' attempted to be set twice in serializer '{}'.".format(key, self._name))
            raise ValueError("A key may only be set once.")
        self._data[key] = value
        self._save(key, value)

        log.debug("Set key {} in serializer".format(key))

    def _open(self):
        """Opens the underlining file."""
        if self._file is None:
            self._file = open(self._filepath, "w", encoding="utf8", newline="\n")

    def _save(self, key, value):
        self._open()
        yaml.dump({key: value}, self._file)


def get_serializer(name="other_data"):
    """
    Generally we want only one Serializer object per kordesii instance to avoid
    clobbering the serialized data each write.

    :param str name: Name of the serializer, should not normally need to be changed.
    :return: Serializer object
    :rtype: Serializer
    """
    global _serializers
    if name not in _serializers:
        # Stored serializer in DECODER_OUTPUT_FILES
        file_path = os.path.join(DECODER_OUTPUT_DIR, name + ".yml")
        _serializers[name] = Serializer(file_path)
    return _serializers[name]


def deserialize(yml_data):
    """
    Deserialize data from the given YAML data.

    :param bytes or str yml_data: Data from the serialization file
    :return: Dict of the deserialized data
    :rtype: dict
    """
    if not yml_data:
        return {}
    try:
        return yaml.load(yml_data)
    except Exception as e:
        log.error("Error loading serialization file: {}".format(str(e)))
        return {}



"""
Contains custom exceptions used to help distinguish emulation errors from
errors caused by bugs.
"""


class FunctionTracingError(Exception):
    def __init__(self, message, ip=None):
        super(FunctionTracingError, self).__init__(message)
        self.ip = ip

    def __str__(self):
        message = super(FunctionTracingError, self).__str__()
        if self.ip is not None:
            return "0x{:X} :: ".format(self.ip) + message
        else:
            return message
